
/* without this #define, fcntl.h won't define O_DIRECT for us */
#define _GNU_SOURCE

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include "odirect.h"

/* If you change this, be sure to keep it a power of 2! */
#define DEFAULT_ALIGN_TO (1<<16)

/* This is actually pretty microsoftish */
#if !defined(O_BINARY)
#	define O_BINARY 0
#endif

/* This isn't :) */
#if !defined(O_LARGEFILE)
#	define O_LARGEFILE 0
#endif

/* Plese note that POINTER_TO_BITWISEABLE, BITWISEABLE_TO_POINTER, is_a_power_of_2 and is_aligned all make some hefty
 * assumptions about the representation of a number that should
 * hold on nearly all contemporary machines, but probably not all machines from now to enternity, nor or some older
 * machines either.  Then again, on those machines, O_DIRECT probably either won't exist, or will have different
 * requirements anyway :)
 */

#define POINTER_TO_BITWISEABLE(pointer) ((TYPE_FOR_BITWISE)pointer)
#define BITWISEABLE_TO_POINTER(bitwiseable) ((char *)bitwiseable)

#if NEVER
char *initodirectcmodule(void)
	{
	}
#endif

/* The algorithm here is to count the number of 1 bits in the word.  If the count is 1, this is a power of two.  Otherwise, it's not :) */
static int odirect_is_a_power_of_2(long number)
	{
	unsigned long current_bit;
	unsigned int number_of_true_bits;
	unsigned int highest_possible_bit;

	/* This could also be calculated in terms of values in limits.h, but this seems to work on systems with or without that header */
	highest_possible_bit = 1 << (sizeof(unsigned long) * 8 - 1);
	
	for (number_of_true_bits = 0, current_bit = highest_possible_bit; current_bit != 0; current_bit = current_bit >> 1)
		{
#if defined(DEBUG)
		fprintf(stderr, "number_of_true_bits is %u, current_bit is %u, number is %u\n", number_of_true_bits, current_bit, number);
#endif
		if ((number & current_bit) != 0)
			{
			number_of_true_bits++;
			}
		}
	if (number_of_true_bits == 1)
		{
		return 1;
		}
	else
		{
		return 0;
		}
	}

static int odirect_is_aligned(char *pointer, int align_to)
	{
	if (odirect_is_a_power_of_2(align_to))
		{
		if ((POINTER_TO_BITWISEABLE(pointer) & (align_to-1)) == 0)
			{
			return 1;
			}
		else
			{
			return 0;
			}
		}
	else
		{
		fprintf(stderr, "Sorry, aligned_to in function aligned must be a power of 2, but instead I got %d\n", align_to);
		exit(1);
		}
	}

void odirect_aligned_malloc(int maximum_buffer_size, int align_to, struct odirect_buffer *odirect_buffer)
	{
	/* char *real_buffer; */
	/* char *aligned_buffer; */

	/* define what power of two boundary we align to.  If we're passed 0, then use a default, otherwise make sure we were passed a power of 2, and use that
	 * instead
	 */
	if (align_to == 0)
		{
		align_to = DEFAULT_ALIGN_TO;
		/* good to go */
		}
	else
		{
		if (odirect_is_a_power_of_2(align_to))
			{
			/* We're good to go */
			}
		else
			{
			odirect_buffer->aligned_buffer = NULL;
			/* Not good */
			return;
			}
		/* printf("%p\n", odirect_buffer->aligned_buffer); */
		/* fflush(stdout); */
		}

	odirect_buffer->align_to = align_to;

	/* this is actually a little bit overly restrictive, but it may help keep things simple - users could, actually, give
	 * us a buffer size that isn't a power of two - we'd just have to be careful to only use the largest power of two in
	 * size that fits inside that buffer (taking into account the alignment requirement, of course).  But that's going to
	 * be prone to resulting in memory waste anyway...
	 */
	if (!odirect_is_a_power_of_2(maximum_buffer_size))
		{
		odirect_buffer->aligned_buffer = NULL;
		return;
		}
	odirect_buffer->maximum_buffer_size = maximum_buffer_size;

	/* here we are -intentionally- allocating -more- memory than the user requested - because we need some elbow room in
	 * which to get our alignment right
	 */
	odirect_buffer->real_buffer = (char *)malloc((odirect_buffer->maximum_buffer_size)+odirect_buffer->align_to-1);
	if (odirect_buffer->real_buffer == NULL)
		{
		odirect_buffer->aligned_buffer = NULL;
		return;
		}

	odirect_buffer->aligned_buffer = odirect_buffer->real_buffer;
	if (odirect_is_aligned(odirect_buffer->aligned_buffer, odirect_buffer->align_to))
		{
		/* we're good to go already, without taking any extra steps  - might happen once in a while :)  malloc() is
		 * supposed to return memory that's aligned for -some- purposes, but not necessarily (and not that likely) -this-
		 * purpose.   ISTR it'll just be aligned for a double or something (or the largest fundamental type), and then
		 * only on architecture that "need" it - indeed, empirically speaking on an Ubuntu 5.10 system with kernel
		 * 2.6.12-9-386 (in case the underlying sbrk() matters) and C library libc-2.3.5.so, the return value of malloc
		 * appears to always have "8" in the least significant nybble.  The other bits seemed pretty chaotic though :)
		 */
		}
	else
		{
		/* set all the relevant low bits to 1, or them in, then add 1.  This should round up to the nearest
		 * suitably-aligned address, where "suitably" in this case means "on a non-tiny power of two"
		 */
		odirect_buffer->aligned_buffer = BITWISEABLE_TO_POINTER((POINTER_TO_BITWISEABLE(odirect_buffer->aligned_buffer) | ((odirect_buffer->align_to)-1))+1);

		/* This shouldn't be necessary - we're testing something that should be an invariant - but if something -does- go wrong, we want to know ASAP, and
		 * binary math can be a "bit" fiddly :-)
		 */
		if (!odirect_is_aligned(odirect_buffer->aligned_buffer, odirect_buffer->align_to))
			{
			fprintf(stderr, "Apologies - something's wrong with libodirect's buffer alignment algorithm and/or test\n");
			exit(1);
			}
		}
	return;
	}

void odirect_aligned_free(struct odirect_buffer *odirect_buffer)
	{
	free(odirect_buffer->real_buffer);
	/* Do -not- free the aligned_buffer!!!!  And don't free the struct either - that's the callers business in this
	 * module.
	 */
	}

struct odirect_buffer *create_odirect_buffer()
	{
	return (struct odirect_buffer *)malloc(sizeof(struct odirect_buffer));
	}

void destroy_odirect_buffer(struct odirect_buffer *odirect_buffer)
	{
	free(odirect_buffer);
	}

int odirect_open(const char *filename, const char *ASCII_flags)
	{
	int flags;
	int retval;
	int file_descriptor;

	/* printf("filename is %s, ASCII_FLAGS are %s\n", filename, ASCII_flags); */
	/* here, we're trying to be like fopen, except we default to having O_DIRECT set.  Also, we can turn off O_DIRECT by passing "d" */
	flags = O_LARGEFILE;
	if (strchr(ASCII_flags, 'r') != NULL)
		{
		if (strchr(ASCII_flags, '+') != NULL)
			{
			flags = flags | O_RDWR;
			}
		else
			{
			flags = flags | O_RDONLY;
			}
		}
	if (strchr(ASCII_flags, 'w') != NULL)
		{
		if (strchr(ASCII_flags, '+') != NULL)
			{
			flags = flags | O_RDWR;
			flags = flags | O_TRUNC;
			}
		else
			{
			flags = flags | O_WRONLY;
			}
		}
	if (strchr(ASCII_flags, 'a') != NULL)
		{
		flags = flags | O_APPEND;
		}
	if (strchr(ASCII_flags, 'b') != NULL)
		{
		flags = flags | O_BINARY;
		}
	if (strchr(ASCII_flags, 'd') != NULL)
		{
		/* If 'd' is specified, we ensure that we use O_DIRECT
		 */
		flags = flags | O_DIRECT;
		}
	else
		{
		/* If 'd' is not passed, do not use O_DIRECT.  This is somewhat important for making performance comparisions
		 * easy.  It shouldn't have been on already, but just to be on the safe side :)
		 */
		flags = flags & (~O_DIRECT);
		}

	/* printf("flags are %d\n", flags); */
	retval = open(filename, flags);
	if (retval < 0)
		{
		return retval;
		}
	else
		{
		file_descriptor = retval;
		}

	return file_descriptor;
	}

struct binary_data odirect_pull(int file_descriptor, struct odirect_buffer *odirect_buffer)
	{
	struct binary_data binary_data;
	int resulting_length;
	
	/* perform the aligned read */
	/* printf("%p\n", odirect_buffer->aligned_buffer); */
	/* fflush(stdout); */
	resulting_length = read(file_descriptor, odirect_buffer->aligned_buffer, odirect_buffer->maximum_buffer_size);
	/* printf("C odirect_pull resulting_length is %d\n", resulting_length); */
	if (resulting_length < 0)
		{
		odirect_buffer->current_buffer_size=0;
		binary_data.size = -1;
		binary_data.data = NULL;
		return binary_data;
		}
	else
		{
		odirect_buffer->current_buffer_size=resulting_length;
		binary_data.size = resulting_length;
		binary_data.data = odirect_buffer->aligned_buffer;
		return binary_data;
		}
	}

int odirect_push(int file_descriptor, struct odirect_buffer *odirect_buffer, struct binary_data binary_data)
	{
	int resulting_length;
	
	/* perform the aligned read */
	/* printf("%p\n", odirect_buffer->aligned_buffer); */
	/* fflush(stdout); */
	if (binary_data.size > odirect_buffer->maximum_buffer_size || binary_data.size % odirect_buffer->align_to != 0)
		{
		return -1;
		}
	else
		{
		odirect_buffer->current_buffer_size = binary_data.size;
		memcpy(odirect_buffer->aligned_buffer, binary_data.data, odirect_buffer->current_buffer_size);
		}
	resulting_length = write(file_descriptor, odirect_buffer->aligned_buffer, odirect_buffer->maximum_buffer_size);
	return resulting_length;
	}

int odirect_close(int file_descriptor)
	{
	int retval;
	retval = close(file_descriptor);
	return retval;
	}

